/**
 * Lecture ``Data Processing on Modern Hardware''
 *
 * Simple SIMD Exercise
 *
 * Based on code by Louis Woods (ETH Zurich, now Oracle Labs)
 * Adapted by Henning Funke ( including techniques V3-6 ) 
 *
 * Compiles with 'gcc -march=native count.c',
 *               'gcc -march=native -O3 -fno-tree-vectorize count.c', or
 *               'gcc -march=native -O3 count.c'
 *  for different levels of auto-vectorization.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <sys/time.h>

#include <immintrin.h>
#include <emmintrin.h>
#include <smmintrin.h>
#include <tmmintrin.h>


#define ARRAYSIZE 1000000000


void print_epi8 ( __m128i v ) {
    printf ( "(%i,%i,%i,%i,%i,%i,%i,%i,%i,%i,%i,%i,%i,%i,%i,%i)\n",
       _mm_extract_epi8 ( v,  0 ),
       _mm_extract_epi8 ( v,  1 ),
       _mm_extract_epi8 ( v,  2 ),
       _mm_extract_epi8 ( v,  3 ),
       _mm_extract_epi8 ( v,  4 ),
       _mm_extract_epi8 ( v,  5 ),
       _mm_extract_epi8 ( v,  6 ),
       _mm_extract_epi8 ( v,  7 ),
       _mm_extract_epi8 ( v,  8 ),
       _mm_extract_epi8 ( v,  9 ),
       _mm_extract_epi8 ( v, 10 ),
       _mm_extract_epi8 ( v, 11 ),
       _mm_extract_epi8 ( v, 12 ),
       _mm_extract_epi8 ( v, 13 ),
       _mm_extract_epi8 ( v, 14 ),
       _mm_extract_epi8 ( v, 15 )
    );
}


void print_epi32 ( __m128i v ) {
    printf ( "(%i,%i,%i,%i)\n",
       _mm_extract_epi32 ( v,  0 ),
       _mm_extract_epi32 ( v,  1 ),
       _mm_extract_epi32 ( v,  2 ),
       _mm_extract_epi32 ( v,  3 )
    );
}

unsigned int count_scalar ( uint8_t* numberarray, size_t size ) {
    unsigned int scalarcnt = 0;
    for (unsigned int i = 0; i < ARRAYSIZE; i++)
    {
        if(numberarray[i] > 42)
            scalarcnt++;
    }
    return scalarcnt;
}


uint32_t count_simd_shuffle ( uint8_t* numberarray, size_t size ) {  
  
    __m128i next4numbers;
    __m128i next16numbers;

    const __m128i comparatorMask = _mm_set_epi32(42,42,42,42);
        __m128i accumulator          = _mm_setzero_si128();

    const __m128i shuffmask1 = _mm_set_epi8 (0xff, 0xff, 0xff, 3,
                                             0xff, 0xff, 0xff, 2,
                                             0xff, 0xff, 0xff, 1,
                                             0xff, 0xff, 0xff, 0);

    const __m128i shuffmask2 = _mm_set_epi8 (0xff, 0xff, 0xff, 7,
                                             0xff, 0xff, 0xff, 6,
                                             0xff, 0xff, 0xff, 5,
                                             0xff, 0xff, 0xff, 4);

    const __m128i shuffmask3 = _mm_set_epi8 (0xff, 0xff, 0xff, 11,
                                             0xff, 0xff, 0xff, 10,
                                             0xff, 0xff, 0xff, 9,
                                             0xff, 0xff, 0xff, 8);

    const __m128i shuffmask4 = _mm_set_epi8 (0xff, 0xff, 0xff, 15,
                                             0xff, 0xff, 0xff, 14,
                                             0xff, 0xff, 0xff, 13,
                                             0xff, 0xff, 0xff, 12);

    for (int i = 0; i < size/16; i++) {
        next16numbers = _mm_loadu_si128((__m128i *) (numberarray+i*16));

        next4numbers = _mm_shuffle_epi8(next16numbers, shuffmask1);
        next4numbers = _mm_cmpgt_epi32(next4numbers, comparatorMask);
        accumulator  = _mm_add_epi32(accumulator, next4numbers);

        next4numbers = _mm_shuffle_epi8(next16numbers, shuffmask2);
        next4numbers = _mm_cmpgt_epi32(next4numbers, comparatorMask);
        accumulator  = _mm_add_epi32(accumulator, next4numbers);

        next4numbers = _mm_shuffle_epi8(next16numbers, shuffmask3);
        next4numbers = _mm_cmpgt_epi32(next4numbers, comparatorMask);
        accumulator  = _mm_add_epi32(accumulator, next4numbers);

        next4numbers = _mm_shuffle_epi8(next16numbers, shuffmask4);
        next4numbers = _mm_cmpgt_epi32(next4numbers, comparatorMask);
        accumulator  = _mm_add_epi32(accumulator, next4numbers);
    }

    uint32_t parallelcnt = _mm_extract_epi32(accumulator,0)+_mm_extract_epi32(accumulator,1)+_mm_extract_epi32(accumulator,2)+_mm_extract_epi32(accumulator,3);
    parallelcnt *= -1;
    return parallelcnt;
}


uint32_t count_simd_reduce ( uint8_t* numberarray, size_t size ) {
    uint32_t parallelcnt = 0;
    __m128i next16numbers;
    const __m128i comparatorMask = _mm_set_epi8(42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42);
    __m128i zero                 = _mm_setzero_si128();
    __m128i accumulator          = _mm_setzero_si128();

    for (int i = 0; i < ARRAYSIZE/16; i++) {
        next16numbers = _mm_loadu_si128 ((__m128i *) (numberarray+i*16));
        next16numbers = _mm_cmpgt_epi8(next16numbers, comparatorMask);
        next16numbers = _mm_abs_epi8(next16numbers);
        next16numbers = _mm_sad_epu8(next16numbers, zero);
        accumulator   = _mm_add_epi64(accumulator, next16numbers);
    }
    parallelcnt   = _mm_extract_epi64(accumulator,0) + _mm_extract_epi64(accumulator,1);
    return parallelcnt;
}

uint32_t count_simd_raking ( uint8_t* numberarray, size_t size ) {

    /* coordination variables */
    unsigned int cursor = 0;
    unsigned int blocksize = 256*16; // =4096
    unsigned int limit;

    /* aggregation variables */
    __m128i constants = _mm_set1_epi8 ( (unsigned char)42 );
    __m128i ones      = _mm_set1_epi8 ( (unsigned char)1 );
    __m128i values, counts;
    unsigned int ssecnt = 0;

    /* parallel count full blocks */
    while ( cursor + blocksize <= ARRAYSIZE ) {

        counts = _mm_setzero_si128();

        for ( limit = cursor + blocksize; cursor < limit; cursor += 16 ) {

            /* load data */
            values = _mm_load_si128 ( (__m128i*) (numberarray + cursor) );

            /* compare and add counts */
            counts = _mm_adds_epu8 ( 
                counts, _mm_and_si128 ( 
                    ones, _mm_cmpgt_epi8 (
                        values, constants 
                    ) 
                )     
            );
        }

        /* aggregate vector register contents */
        ssecnt += _mm_extract_epi8 ( counts,  0 );
        ssecnt += _mm_extract_epi8 ( counts,  1 );
        ssecnt += _mm_extract_epi8 ( counts,  2 );
        ssecnt += _mm_extract_epi8 ( counts,  3 );
        ssecnt += _mm_extract_epi8 ( counts,  4 );
        ssecnt += _mm_extract_epi8 ( counts,  5 );
        ssecnt += _mm_extract_epi8 ( counts,  6 );
        ssecnt += _mm_extract_epi8 ( counts,  7 );
        ssecnt += _mm_extract_epi8 ( counts,  8 );
        ssecnt += _mm_extract_epi8 ( counts,  9 );
        ssecnt += _mm_extract_epi8 ( counts, 10 );
        ssecnt += _mm_extract_epi8 ( counts, 11 );
        ssecnt += _mm_extract_epi8 ( counts, 12 );
        ssecnt += _mm_extract_epi8 ( counts, 13 );
        ssecnt += _mm_extract_epi8 ( counts, 14 );
        ssecnt += _mm_extract_epi8 ( counts, 15 );
    }

    /* scalar count remaining values */
    for ( ; cursor < ARRAYSIZE; cursor++ ) {
        if(numberarray[cursor] > 42)
            ssecnt++;
    }
    return ssecnt;
}

uint32_t count_simd_raking_signtrick ( uint8_t* numberarray, size_t size ) {

    /* coordination variables */
    unsigned int cursor = 0;
    unsigned int blocksize = 128*16; // =2048
    unsigned int limit;

    /* aggregation variables */
    __m128i constants = _mm_set1_epi8 ( (unsigned char)42 );
    __m128i values, counts;
    unsigned int ssecnt = 0;

    /* parallel count full blocks */
    while ( cursor + blocksize <= ARRAYSIZE ) {

        counts = _mm_setzero_si128();

        for ( limit = cursor + blocksize; cursor < limit; cursor += 16 ) {

            /* load data */
            values = _mm_load_si128 ( (__m128i*) (numberarray + cursor) );

            /* compare and add counts */
            counts = _mm_sub_epi8 ( 
                counts, _mm_cmpgt_epi8 (
                    values, constants 
                ) 
            );
        }

        /* aggregate vector register contents */
        ssecnt += _mm_extract_epi8 ( counts,  0 );
        ssecnt += _mm_extract_epi8 ( counts,  1 );
        ssecnt += _mm_extract_epi8 ( counts,  2 );
        ssecnt += _mm_extract_epi8 ( counts,  3 );
        ssecnt += _mm_extract_epi8 ( counts,  4 );
        ssecnt += _mm_extract_epi8 ( counts,  5 );
        ssecnt += _mm_extract_epi8 ( counts,  6 );
        ssecnt += _mm_extract_epi8 ( counts,  7 );
        ssecnt += _mm_extract_epi8 ( counts,  8 );
        ssecnt += _mm_extract_epi8 ( counts,  9 );
        ssecnt += _mm_extract_epi8 ( counts, 10 );
        ssecnt += _mm_extract_epi8 ( counts, 11 );
        ssecnt += _mm_extract_epi8 ( counts, 12 );
        ssecnt += _mm_extract_epi8 ( counts, 13 );
        ssecnt += _mm_extract_epi8 ( counts, 14 );
        ssecnt += _mm_extract_epi8 ( counts, 15 );
    }

    /* scalar count remaining values */
    for ( ; cursor < ARRAYSIZE; cursor++ ) {
        if(numberarray[cursor] > 42)
            ssecnt++;
    }
    return ssecnt;
}

uint32_t count_simd_raking_opt ( uint8_t* numberarray, size_t size ) {
    /* coordination variables */
    unsigned int cursor = 0;
    unsigned int blocksize = 128*16; // =2048
    unsigned int limit;

    /* aggregation variables */
    __m128i constants      = _mm_set1_epi8 ( (unsigned char) 42 );
    __m128i zero           = _mm_setzero_si128();
    __m128i values, counts;
    __m128i accumulator          = _mm_setzero_si128();
    unsigned int ssecnt = 0;

    /* parallel count full blocks */
    while ( cursor + blocksize <= ARRAYSIZE ) {

        counts = _mm_setzero_si128();

        for ( limit = cursor + blocksize; cursor < limit; cursor += 16 ) {

            /* load data */
            values = _mm_load_si128 ( (__m128i*) (numberarray + cursor) );

            /* compare and add counts */
            counts = _mm_sub_epi8 ( 
                counts, _mm_cmpgt_epi8 (
                    values, constants 
                ) 
            );
        }
        /* reduce by Sum of Absolute Differences */
        __m128i reduced = _mm_sad_epu8 ( counts, zero);
        accumulator   = _mm_add_epi64 ( accumulator, reduced );
    }
    ssecnt   = _mm_extract_epi64(accumulator,0) + _mm_extract_epi64(accumulator,1);
    /* scalar count remaining values */
    for ( ; cursor < ARRAYSIZE; cursor++ ) {
        if(numberarray[cursor] > 42)
            ssecnt++;
    }
    return ssecnt;
}


#ifdef __AVX512BW__
uint32_t count_simd_avx512 ( uint8_t* numberarray, size_t size ) {

    /* coordination variables */
    unsigned int cursor = 0;

    /* aggregation variables */
    __m512i constants = _mm512_set1_epi8 ( (unsigned char) 42 );
    __m512i values;
    __mmask64 mask;
    unsigned int accumulator = 0;

    /* parallel count full blocks */
    while ( cursor <= ARRAYSIZE - 64 ) {
        values = _mm512_loadu_si512 ( numberarray + cursor );
        mask = _mm512_cmpgt_epi8_mask ( values, constants );
        accumulator += _mm_popcnt_u64 ( mask );
        cursor += 64;
    }
    /* scalar count remaining values */
    for ( ; cursor < ARRAYSIZE; cursor++ ) {
        if(numberarray[cursor] > 42)
            accumulator++;
    }
    return accumulator;
}
#endif

unsigned long get_ms ( struct timeval start, struct timeval end ) {
    return ( (end.tv_sec - start.tv_sec) * 1000000 + (end.tv_usec - start.tv_usec) ) / 1000;
}

void printStats ( unsigned int res, unsigned long time, unsigned long timeScalar ) {
    printf ( "result: %u\n", res ); 
    printf ( "time: %lu ms\n", time );
    printf ( "throughput: %.1f GB/s\n", ( (double)ARRAYSIZE / 1000000000.0 ) / ( (double)time / 1000.0 ) );
    if ( timeScalar != 0 )
        printf ( "speedup: %.1fx\n", (double) timeScalar / (double) time );
    printf ( "\n" );
}


int main (int argc, char **argv)
{


    const __m128i a = _mm_setr_epi8 ( 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 );
    const __m128i b = _mm_set_epi8 (0xff, 0xff, 4, 3,
                                    0xff, 0xff, 3, 2,
                                    0xff, 0xff, 2, 1,
                                    0xff, 2, 1, 0);

    print_epi8 ( a );
    print_epi8 ( b );

    __m128i c = _mm_shuffle_epi8 ( a, b );

    print_epi8 ( c );
    
    print_epi32 ( c );

    return 0;

    struct timeval start, end;
    long timeScalar;
    

    /* allocate array of 8-bit integers */
    uint8_t *numberarray = malloc (ARRAYSIZE * sizeof (*numberarray));
    if (numberarray == NULL)
    {
        fprintf (stderr, "error during memory allocation\n");
        exit(EXIT_FAILURE);
    }

    /* populate array with random numbers */
    for (unsigned int i = 0; i < ARRAYSIZE; i++)
    {
        numberarray[i] = rand () % 100;
    }

    unsigned int res;


    printf ( "  ---  scalar count (1 value at a time) ---  \n" );
    gettimeofday (&start, NULL);
    res = count_scalar ( numberarray, ARRAYSIZE );
    gettimeofday (&end, NULL);
    timeScalar = get_ms ( start, end );
    printStats ( res, timeScalar, 0 );

    printf ( "  ---  SIMD count V1: Shuffle-based (4 vals at a time)  ---  \n" );
    gettimeofday (&start, NULL);
    res = count_simd_shuffle ( numberarray, ARRAYSIZE );
    gettimeofday (&end, NULL);
    printStats ( res, get_ms ( start, end ), timeScalar );

    printf ( "  ---  SIMD count V2: Reduction by Sum-of-Absolute-Diffs (16 vals at a time)  ---  \n" );
    gettimeofday (&start, NULL);
    res = count_simd_reduce ( numberarray, ARRAYSIZE );
    gettimeofday (&end, NULL);
    printStats ( res, get_ms ( start, end ), timeScalar );

    printf ( "  ---  SIMD count V3: Raking (16 values at a time)  ---  \n" );
    gettimeofday (&start, NULL);
    res = count_simd_raking ( numberarray, ARRAYSIZE );
    gettimeofday (&end, NULL);
    printStats ( res, get_ms ( start, end ), timeScalar );

    printf ( "  ---  SIMD count V4: Raking with sign trick (16 vals at a time)  ---  \n" );
    gettimeofday (&start, NULL);
    res = count_simd_raking_signtrick ( numberarray, ARRAYSIZE );
    gettimeofday (&end, NULL);
    printStats ( res, get_ms ( start, end ), timeScalar );

    printf ( "  ---  SIMD count V5: Raking opt (16 vals at a time)  ---  \n" );
    gettimeofday (&start, NULL);
    res = count_simd_raking_opt ( numberarray, ARRAYSIZE );
    gettimeofday (&end, NULL);
    printStats ( res, get_ms ( start, end ), timeScalar );

    #ifdef __AVX512BW__
    printf ( "  ---  SIMD count V6: AVX 512  (64 vals at a time)  ---  \n" );
    gettimeofday (&start, NULL);
    res = count_simd_avx512 ( numberarray, ARRAYSIZE );
    gettimeofday (&end, NULL);
    printStats ( res, get_ms ( start, end ), timeScalar );
    #endif


    return EXIT_SUCCESS;
}


